/*-
 * Copyright (C)2018..2023 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2018..2023 @BABOLO http://www.babolo.ru/"
#ident "@(#) $Id: csv2pg.c,v 1.24 2024/12/22 19:39:29 babolo Exp $"

#define MIFE_COMPAT      5
#define BLIN_COMPAT      4
#define Bpars_COMPAT     4

#include <sys/types.h>
#include <sys/event.h>
#include <sys/time.h>
#include <sysexits.h>
#include <stdlib.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include <mife.h>
#include "csv2pg.h"

#define Cn 0x00000800 /* string num     */
#define Ce 0x00001000 /* write escaped  */
#define Cq 0x00002000 /* write quote    */
#define Co 0x00004000 /* write ordinar  */
#define Cd 0x00008000 /* write repeat   */
#define Cf 0x00010000 /* fields control */
#define Ct 0x00020000 /* tuple ready    */
#define Cl 0x00040000 /* eol test       */
#define Cc 0x00080000 /* do cortege     */
#define Cx 0x00800000 /* error          */
#define C_state  0xFF /* next state     */

static const char statenames[][3] = { "bg", "po", "pq", "qq", "qo", "el", "px" };
static const char clasnames[][3] = { "ce", "cc", "cq", "cl", "co", "cd", "xx" };
static const char subs[16] = {'0', 0, 0, 0 , 0, 0, 0, 0 , 'b', 't', 'n', 'v' , 'f', 'r', 0, 0};

enum states
{ bg  /*         */
, po  /*        */
, pq  /*         */
, qq  /*               */
, qo  /*    */
, el  /*           */
, px
};

enum classes
{ ce  /* quoted  */
, cc  /* comma   */
, cq  /* quot    */
, cl  /* eol     */
, co  /* ordinar */
, cd  /* double  */
, xx
};

static const char pn[][3] = 
{ "- ", "- ", "- ", "- ", "- ", "- ", "- ", "- "
, "x ", "- ", "- ", "- ", "c ", "l ", "t ", "f "
, "d ", "o ", "q ", "e ", "n ", "- ", "- ", "- "
};

static enum classes class[256] =
{ ce, co, co, co , co, co, co, co , ce, ce, cl, cl , cl, cl, co, co
, co, co, co, co , co, co, co, co , co, co, co, co , co, co, co, co
, co, co, co, co , co, co, co, co , co, co, co, co , co, co, co, co
, co, co, co, co , co, co, co, co , co, co, co, co , co, co, co, co

, co, co, co, co , co, co, co, co , co, co, co, co , co, co, co, co
, co, co, co, co , co, co, co, co , co, co, co, co , cd, co, co, co
, co, co, co, co , co, co, co, co , co, co, co, co , co, co, co, co
, co, co, co, co , co, co, co, co , co, co, co, co , co, co, co, co

, co, co, co, co , co, co, co, co , co, co, co, co , co, co, co, co
, co, co, co, co , co, co, co, co , co, co, co, co , co, co, co, co
, co, co, co, co , co, co, co, co , co, co, co, co , co, co, co, co
, co, co, co, co , co, co, co, co , co, co, co, co , co, co, co, co

, co, co, co, co , co, co, co, co , co, co, co, co , co, co, co, co
, co, co, co, co , co, co, co, co , co, co, co, co , co, co, co, co
, co, co, co, co , co, co, co, co , co, co, co, co , co, co, co, co
, co, co, co, co , co, co, co, co , co, co, co, co , co, co, co, co
};

static const u_int32_t automa[4][px][xx] =
/*          ce        cc        cq           cl        co        cd     xx */
{ { {    Ce|po,    Cf|bg,       pq,    Ct|Cl|el,    Co|po,    Co|Cd|po }/* bg         */
  , {    Ce|po,    Cf|bg,       qo,    Ct|Cl|el,    Co|po,    Co|Cd|po }/* po        */
  , {    Ce|pq,    Co|pq,       qq,    Ce   |pq,    Co|pq,    Co|Cd|pq }/* pq         */
  , { Cq|Ce|po,    Cf|bg,    Co|pq,    Ct|Cl|el, Cq|Co|po, Cq|Co|Cd|po }/* qq               */
  , {    Cx|px,    Cx|px,    Co|po,    Ct|Cl|el,    Cx|px,       Cx|px }/* qo    */
  , {    Ce|po,    Cf|bg,       pq,       Cl|el,    Co|po,    Co|Cd|po }/* el           */
  }                                                                     /*      ****/
, { {    Ce|po,    Cf|bg,       pq,    Ct|Cl|el,    Co|po,    Co|Cd|po }/* bg         */
  , {    Ce|po,    Cf|bg,       qo,    Ct|Cl|el,    Co|po,    Co|Cd|po }/* po        */
  , {    Ce|pq,    Co|pq,       qq,    Ce   |pq,    Co|pq,    Co|Cd|pq }/* pq         */
  , { Cq|Ce|po,    Cf|bg,    Co|pq,    Ct|Cl|el, Cq|Co|po, Cq|Co|Cd|po }/* qq               */
  , { Cq|Ce|po, Cq|Cf|bg, Cq|Co|po, Cq|Ct|Cl|el, Cq|Co|po, Cq|Co|Cd|po }/* qo    */
  , {    Ce|po,    Cf|bg,       pq,       Cl|el,    Co|po,    Co|Cd|po }/* el           */
  }                                                                     /* -r                   ****/
, { {    Ce|po,    Cf|bg,       pq,    Ct|Cl|el,    Co|po,    Co|Cd|po }/* bg         */
  , {    Ce|po,    Cf|bg,    Cx|px,    Ct|Cl|el,    Co|po,    Co|Cd|po }/* po        */
  , {    Ce|pq,    Co|pq,       qq,    Ce   |pq,    Co|pq,    Co|Cd|pq }/* pq         */
  , {    Cx|px,    Cf|bg,    Co|pq,    Ct|Cl|el,    Cx|px,       Cx|px }/* qq               */
  , {    Cx|px,    Cx|px,    Cx|px,       Cx|px,    Cx|px,       Cx|px }/* qo    */
  , {    Ce|po,    Cf|bg,    Co|po,       Cl|el,    Co|po,    Co|Cd|po }/* el           */
  }                                                                     /*     ****/
, { {    Ce|po,    Cf|bg,       pq,    Ct|Cl|el,    Co|po,    Co|Cd|po }/* bg         */
  , {    Ce|po,    Cf|bg,    Cx|px,    Ct|Cl|el,    Co|po,    Co|Cd|po }/* po        */
  , {    Ce|pq,    Co|pq,       qq,    Ce   |pq,    Co|pq,    Co|Cd|pq }/* pq         */
  , {    Cx|px,    Cf|bg,    Co|pq,    Ct|Cl|el,    Cx|px,       Cx|px }/* qq               */
  , {    Cx|px,    Cx|px,    Cx|px,       Cx|px,    Cx|px,       Cx|px }/* qo    */
  , {    Ce|po,    Cf|bg,    Co|pq,       Cl|el,    Co|po,    Co|Cd|po }/* el           */
  }                                                                     /*      ****/
};

static int
csv2pg_do(csv2pg_cf *cf) {
    blin_statectl    sctl = {{4, 1, 8, 24, ' ', 3, 1, 0, 0}};      /*  blin_stateheader()         */
    int              fieldcount;
    u_int32_t        control;
    u_int32_t        shifter;
    off_t            offset;
    enum states      state;
    u_int32_t        lnctr;
    u_int32_t        line = 0;
    ssize_t          reed;
    u_int32_t        beg  = Cn;
    int              ex   = EX_OK;
    char             ccl  = cl;
    char            *p    = NULL;

#   define blin_internal_flags (cf->flags & BLIN_MASK)
    class[cf->quote] = cq;
    class[cf->comma] = cc;
    if  (cf->flags & CVS2PG_XHEAD) cf->flags |= CVS2PG_THROW;
    blin_stateheader(sctl, pn);
    for ( cf->tuplecount = 0
        , fieldcount = 0
        , control = 0
        , shifter = !!(cf->flags & CVS2PG_RELAX)
        , offset = 0
        , state = bg
        , lnctr = 0
        , reed = 1
        ; (state < px) && !!reed
        ; ++offset
        ) {
        if  (0 > (reed = mife_read(cf->mdi, 1, offset))) {
            ifBLIN_QW0("mife_read");
            ex = EX_IOERR;
            goto out;
        }
        if  (!reed) {
            control = automa[shifter][state][cl];
            p = &ccl;
        } else if (!(p = mife_pointer(cf->mdi))) {
            ifBLIN_QW0("mife_pointer");
            ex = EX_IOERR;
            goto out;
        } else {
            control = automa[shifter][state][class[*p & 0xFF]] | beg;
            beg = 0;
        }
        blin_statebody( sctl
                      , pn
                      , statenames
                      , clasnames[class[*p & 0xFF]]
                      , p
                      , (size_t)offset
                      , control
                      , state
                      , shifter
                      , (u_char)p[0]
                      );
        if  (!!shifter) shifter = 1;
        if  ((control & Cn) && (cf->flags & CVS2PG_NUMBE)) {
            char b[CVS2PG_BUFSZ];

            if  (0 > (ex = snprintf(b, CVS2PG_BUFSZ, "%u", cf->start + line++))) {
                ifBLIN_QW0("snprintf %u", cf->start + line);
                ex = EX_IOERR;
                goto out;
            }
            if  (!ex) {
                ifBLIN_QX0("snprintf = 0");
                ex = EX_SOFTWARE;
                goto out;
            }
            if  (ex >= CVS2PG_BUFSZ) {
                ifBLIN_QX0("snprintf overflow");
                ex = EX_DATAERR;
                goto out;
            }
            ex = EX_OK;
            for (int i = 0; (i < CVS2PG_BUFSZ) && !!b[i]; i++) {
                if  (0 > mife_aput(cf->out, b[i])) {
                    ifBLIN_QW0("mife_aput N%u", i);
                    ex = EX_IOERR;
                    goto out;
            }   }
            if  (0 > mife_aput(cf->out, '\t')) {
                ifBLIN_QW0("mife_aput \t");
                ex = EX_IOERR;
                goto out;
        }   }
        if  (!(cf->flags & (CVS2PG_THCUT | CVS2PG_THROW))) {
            if  (control & Ce) {
                if  (0 > mife_aput(cf->out, '\\')) {
                    ifBLIN_QW0("mife_aput \\");
                    ex = EX_IOERR;
                    goto out;
                }
                if  (0 > mife_aput(cf->out, subs[*p & 0xFF])) {
                    ifBLIN_QW0("mife_aput");
                    ex = EX_IOERR;
                    goto out;
            }   }
            if  (control & Cq) {
                if  (0 > mife_aput(cf->out, cf->quote)) {
                    ifBLIN_QW0("mife_aput");
                    ex = EX_IOERR;
                    goto out;
            }   }
            if  (control & Co) {
                if  (0 > mife_aput(cf->out, *p)) {
                    ifBLIN_QW0("mife_aput");
                    ex = EX_IOERR;
                    goto out;
            }   }
            if  (control & Cd) {
                if  (0 > mife_aput(cf->out, *p)) {
                    ifBLIN_QW0("mife_aput");
                    ex = EX_IOERR;
                    goto out;
        }   }   }
        if  (control & Cf) {
            ++fieldcount;
            if  ((cf->fieldnom >= 0) && (fieldcount >= cf->fieldnom)) {
                if  (cf->flags & (CVS2PG_OVEVA | CVS2PG_UNEVA)) {
                    ifBLIN_QX1( "String %d: field %d ,  %d  XXXX"
                              , cf->tuplecount
                              , fieldcount
                              , cf->fieldnom
                              );
                    if  (cf->flags & CVS2PG_OVEVA) cf->flags |= CVS2PG_THCUT;
                } else {
                    ifBLIN_QX0( "String %d: field %d ,  %d  XXXX"
                              , cf->tuplecount
                              , fieldcount
                              , cf->fieldnom
                              );
                    ex = EX_DATAERR;
                    errno = ERANGE;
                    goto out;
            }   }
            if  ((!(cf->flags & (CVS2PG_THCUT | CVS2PG_THROW))) && (0 > mife_aput(cf->out, '	'))) {
                ifBLIN_QW0("mife_aput");
                ex = EX_IOERR;
                goto out;
        }   }
        if  (control & Ct) {
            if  (!!cf->hdrsize) cf->hdrsize--;
            if  (cf->fieldnom < 0) {
                cf->fieldnom = fieldcount + 1;
                control |= Cc;
            } else if ((fieldcount + 1) >= cf->fieldnom) {
                control |= Cc;
            } else if (cf->flags & CVS2PG_SCEVA) {
                if  (!!shifter && (fieldcount < cf->fieldnom)) {
                    if  (state == po) shifter = 2;
                    if  (state == qq) shifter = 3;
                }
            } else if (cf->flags & CVS2PG_OVEVA) {
                while (!(cf->flags & (CVS2PG_THCUT | CVS2PG_THROW)) && (fieldcount < cf->fieldnom)) {
                    if  (++fieldcount >= cf->fieldnom) break;
                    if  (0 > mife_aput(cf->out, '	')) {
                        ifBLIN_QW0("mife_aput");
                        ex = EX_IOERR;
                        goto out;
                    }
                    if  (0 > mife_aput(cf->out, '\\')) {
                        ifBLIN_QW0("mife_aput \\");
                        ex = EX_IOERR;
                        goto out;
                    }
                    if  (0 > mife_aput(cf->out, 'N')) {
                        ifBLIN_QW0("mife_aput");
                        ex = EX_IOERR;
                        goto out;
                }   }
                control |= Cc;
            } else if (++fieldcount < cf->fieldnom) {
                if  (!(cf->flags & CVS2PG_UNEVA)) {
                    ifBLIN_QX0( "short tuple %d: last field %d of %d"
                              , cf->tuplecount
                              , fieldcount
                              , cf->fieldnom
                              );
                    ex = EX_DATAERR;
                    errno = ERANGE;
                    goto out;
                } else {
                    ifBLIN_QX1( "short tuple %d: last field %d of %d"
                              , cf->tuplecount
                              , fieldcount
                              , cf->fieldnom
                              );
                }
                control |= Cc;
            }
            if  (control & Cc) {
                if  (!(cf->flags & CVS2PG_THROW)) cf->tuplecount++;
                fieldcount = 0;
                lnctr = 0;
                cf->flags &= ~CVS2PG_THCUT;
                if  (!(cf->flags & (CVS2PG_THCUT | CVS2PG_THROW))) {
                    beg  = Cn;
                    if  (0 > mife_aput(cf->out, '\n')) {
                        ifBLIN_QW0("mife_aput");
                        ex = EX_IOERR;
                        goto out;
            }   }   }
            if  (!cf->hdrsize) cf->flags &= ~CVS2PG_THROW;
        }
        if  (control & Cl) {
            if  (!cf->hdrsize && (cf->flags & CVS2PG_OHEAD)) break;
            if  ((1 << (*p & 0xFF)) & lnctr) {
                ifBLIN_QX1("Empty or short string N %d", cf->tuplecount);
            }
            lnctr |= 1 << (*p & 0xFF);
        }
        if  (control & Cx) {
            ifBLIN_QX0("Syntax error at line %d field %d", cf->tuplecount, fieldcount);
            goto out;
        }
        state = control & C_state;
    }
    if  (!ex && (cf->flags & CVS2PG_COEOF)) {
        if  (0 > mife_aput(cf->out, '\\')) {
            ifBLIN_QW0("mife_aput");
            ex = EX_IOERR;
            goto out;
        }
        if  (0 > mife_aput(cf->out, '.')) {
            ifBLIN_QW0("mife_aput");
            ex = EX_IOERR;
            goto out;
        }
        if  (0 > mife_aput(cf->out, '\n')) {
            ifBLIN_QW0("mife_aput");
            ex = EX_IOERR;
            goto out;
    }   }
out:
    if  (0 > mife_aclose(cf->out)) {
        ifBLIN_QW0("mife_aclose");
        ex = EX_IOERR;
    }
    return(ex);
#   undef blin_internal_flags
}

static void
csv2pg_dumpopts(csv2pg_cf *cf) {
#   define blin_internal_flags (cf->flags & BLIN_MASK)
    printf("       quote	-w%c\n", cf->quote);
    printf("       quote	-W%02X\n", cf->quote);
    printf("       comma	-c%c\n", cf->comma);
    printf("       comma	-C%02X\n", cf->comma);
    if  (cf->flags & CVS2PG_NUMBE) {
        printf("       start	-i%u\n", cf->start);
    } else {
        printf("       start	  %u\n", cf->start);
    }
    printf("   increment	  %u\n", cf->increment);
    if  (cf->fieldnom >= 0)        printf("   fieldsnum	-n%d\n", cf->fieldnom);
    if  (cf->inname)               printf("       input	-f%s\n", cf->inname);
    if  (cf->outname)              printf("      output	-o%s\n", cf->outname);
    if  (cf->flags & CVS2PG_UNEVA) printf("    unallign	-u\n");
    if  (cf->flags & CVS2PG_OVEVA) printf("      allign	-x\n");
    if  (cf->flags & CVS2PG_COEOF) printf("         eof	-e\n");
    if  (cf->flags & CVS2PG_RELAX) printf("  relaxquote	-r\n");
    if  (cf->flags & CVS2PG_WHENY) printf("     sizeout	-N\n");
    if  (cf->flags & CVS2PG_XHEAD) printf("     headcut	-h\n");
    if  (cf->flags & CVS2PG_OHEAD) printf("    headonly	-H\n");
    if  (cf->flags & CVS2PG_SCEVA) printf("        glue	-g\n");
    return;
#   undef blin_internal_flags
}

static void __attribute__((__noreturn__))
usage(int ex) {
    fprintf( stderr
           , "csv2pg @BABOLO V.M "VERS"  "DATE" http://www.babolo.ru/\n"
             "    -b NUMBER - header size (default 1)\n"
             "    -c SYM    - comma symbol literal (default ,)\n"
             "    -C CODE   - comma symbol code (default 2C)\n"
             "    -e        - \\. at end\n"
             "    -f FILE   - from file, default stdin\n"
             "    -g        - glue short lines together\n"
             "    -h        - header delete\n"
             "    -H        - header only\n"
             "    -i NUMBER - create first column with string numbers, start with NUMBER\n"
             "    -n NUMBER - number of fields\n"
             "    -N        - print number of fields and tuples\n"
             "    -o FILE   - to file, default stdout\n"
             "    -q        - quiet\n"
             "    -r        - allow single quote inside field\n"
             "    -u        - get fields number as is\n"
             "    -v        - verbose\n"
             "    -w SYM    - quote symbol literal (default \")\n"
             "    -W CODE   - quote symbol code (default 5C)\n"
             "    -x        - align field number\n"
             "    -?        - this text\n"
           );
    exit(ex);
}

int
main(int argc, char *argv[]) {
    const char      *flar = "b:c:C:ef:ghHi:n:No:qrsuvw:W:x=:?";
    babolo_opts     *bos = NULL;
    int              tfd;
    int              ex = EX_OK;
    csv2pg_cf       *cf;
    const char      *o;
    int              c;

    if  (!(cf = calloc(1, sizeof(csv2pg_cf)))) {
        ifBLIN_QW0("No mem %"BLIN_X, sizeof(csv2pg_cf));
        errno = ENOMEM;
        ex = EX_OSERR;
        goto out;
    }
    cf->flags = BLIN_BIT0;
#   define blin_internal_flags (cf->flags & BLIN_MASK)
    cf->fieldnom = -1;
    cf->hdrsize = 1;
    cf->quote = '\"';
    cf->comma = ',';
    cf->start = 1;
    cf->increment = 1;
    if  (!(bos = babolo_openopts(BLIN_BIT0 | Bpars_NONU | Bpars_SYME | Bpars_ESYM, 'Z'))) {
        ifBLIN_QW0("babolo_openopts");
        ex = EX_SOFTWARE;
        goto out;
    }
    if  (babolo_setopts(bos, Bpars_NONU, argc, argv, flar)) {
        ifBLIN_QW0("babolo_setopts");
        ex = EX_SOFTWARE;
        goto out;
    }
    while (0 < (c = babolo_getopts(bos))) {
        switch (c) {
        case 'b': if  (1 != cf->hdrsize) ifBLIN_QX1("Header size redefined");
                ; errno = 0;
                ; cf->hdrsize = (u_int16_t)babolo_getoptunum(bos, 0);
                ; if  (errno) {
                ;     ifBLIN_QW0("-%c babolo_getoptunum", bos->c);
                ;     ex = EX_USAGE;
                ;     goto out;
                ; }
                ; break;
        case 'c': if  (cf->comma != ',')  ifBLIN_QX1("Comma redefined");
                ; if  (!(o = babolo_getoptsarg(bos))) {
                ;     ifBLIN_QW0("No -%c arg", bos->c);
                ;     ex = EX_USAGE;
                ;     goto out;
                ; }

                ; cf->comma = (u_char)*o;
                ; break;
        case 'C': if  (cf->comma != ',')  ifBLIN_QX1("Comma redefined");
                ; errno = 0;
                ; cf->comma = (u_char)babolo_getoptunum(bos, 0);
                ; if  (errno) {
                ;     ifBLIN_QW0("-%c babolo_getoptunum", bos->c);
                ;     ex = EX_USAGE;
                ;     goto out;
                ; }
                ; break;
        case 'e': cf->flags |= CVS2PG_COEOF;
                ; break;
        case 'f': if  (!!cf->mdi) ifBLIN_QX1("Input file redefined");
                ; if  (!(o = babolo_getoptsarg(bos))) {
                ;     ifBLIN_QW0("No -%c arg", bos->c);
                ;     ex = EX_USAGE;
                ;     goto out;
                ; }
                ; if  (!(cf->mdi = mife_init(BLIN_BIT0 | MIFE_SLID))) {
                ;     ifBLIN_QW0("mife_init");
                ;     ex = EX_SOFTWARE;
                ;     goto out;
                ; }
                ; if  (0 > mife_ctlfile(cf->mdi, o)) {
                ;     ifBLIN_QW0("mife_ctlfile %s", o);
                ;     ex = EX_IOERR;
                ;     goto out;
                ; }
                ; cf->inname = o;
                ; break;
        case 'g': cf->flags |= CVS2PG_SCEVA;
                ; break;
        case 'h': cf->flags |= CVS2PG_XHEAD;
                ; break;
        case 'H': cf->flags |= CVS2PG_OHEAD;
                ; break;
        case 'i': if  (cf->flags & CVS2PG_NUMBE) ifBLIN_QX1("Start string number redefined");
                ; cf->flags |= CVS2PG_NUMBE;
                ; errno = 0;
                ; cf->start = (u_int16_t)babolo_getoptunum(bos, 0);
                ; if  (errno) {
                ;     ifBLIN_QW0("-%c babolo_getoptunum", bos->c);
                ;     ex = EX_USAGE;
                ;     goto out;
                ; }
                ; break;
        case 'n': if  (0 < cf->fieldnom) ifBLIN_QX1("Field quantity redefined");
                ; errno = 0;
                ; cf->fieldnom = (int)babolo_getoptsnum(bos, 0);
                ; if  (errno) {
                ;     ifBLIN_QW0("-%c babolo_getoptsnum", bos->c);
                ;     ex = EX_USAGE;
                ;     goto out;
                ; }
                ; break;
        case 'N': cf->flags |= CVS2PG_WHENY;
                ; break;
        case 'o': if  (!!cf->out) ifBLIN_QX1("Output file redefined");
                ; if  (!(o = babolo_getoptsarg(bos))) {
                ;     ifBLIN_QW0("No -%c arg", bos->c);
                ;     ex = EX_USAGE;
                ;     goto out;
                ; }
                ; if  (0 > (tfd = open(o, O_WRONLY | O_CREAT | O_TRUNC, 0666))) {
                ;     ifBLIN_QW0("open %s", o);
                ;     ex = EX_IOERR;
                ;     goto out;
                ; }
                ; if  (!(cf->out = mife_aopen(BLIN_BIT0, tfd))) {
                ;     ifBLIN_QW0("mife_aopen %s", o);
                ;     ex = EX_IOERR;
                ;     goto out;
                ; }
                ; cf->outname = o;
                ; break;
        case 'q': cf->flags &= ~BLIN_MASK;
                ; break;
        case 'r': cf->flags |= CVS2PG_RELAX;
                ; break;
        case 's': break; /*     */
        case 'u': cf->flags |= CVS2PG_UNEVA;
                ; break;
        case 'v': BLIN_VERBOSE(cf->flags);
                ; break;
        case 'w': if  (cf->quote != '\"')  ifBLIN_QX1("Quote redefined");
                ; if  (!(o = babolo_getoptsarg(bos))) {
                ;     ifBLIN_QW0("No -%c arg", bos->c);
                ;     ex = EX_USAGE;
                ;     goto out;
                ; }
                ; cf->quote = (u_char)*o;
                ; break;
        case 'W': if  (cf->quote != '\"')  ifBLIN_QX1("Quote redefined");
                ; errno = 0;
                ; cf->quote = (u_char)babolo_getoptunum(bos, 0);
                ; if  (errno) {
                ;     ifBLIN_QW0("-%c babolo_getoptunum", bos->c);
                ;     ex = EX_USAGE;
                ;     goto out;
                ; }
                ; break;
        case 'x': cf->flags |= CVS2PG_OVEVA;
                ; break;
        case '=': if  (!(o = babolo_getoptsarg(bos))) {
                ;     ifBLIN_QW0("No -%c arg", bos->c);
                ;     ex = EX_USAGE;
                ;     goto out;
                ; }
                ; switch(o[0]) {
                  case '0': blin_ctl(BLIN_CTL_DUMP);           exit(EX_OK);
                  case '1': babolo_optest(flar);               exit(EX_OK);
                  case '2': babolo_optext(bos);                exit(EX_OK);
                  case '3': babolo_dumpopts(bos);              exit(EX_OK);
                  case '=': csv2pg_dumpopts(cf);               exit(EX_OK);
                  }
                ; /* FALLTHROUGH */
        case '?': usage(EX_OK);
        default : ifBLIN_QW0("Ill flag -%c", bos->c);
                ; usage(EX_USAGE);
    }   }
    if  (0 > c) usage(EX_USAGE);
    if  (!cf->mdi) {
        if  (!(cf->mdi = mife_init(MIFE_SLID))) {
            ifBLIN_QW0("mife_init");
            ex = EX_IOERR;
            goto out;
        }
        if  (0 > mife_ctlfdsc(cf->mdi, fileno(stdin))) {
            ifBLIN_QW0("mife_ctlfdsc stdin");
            ex = EX_IOERR;
            goto out;
        }
        cf->inname = "/dev/stdin";
    }
    if  (!cf->out) {
        if  (!(cf->out = mife_aopen(0, fileno(stdout)))) {
            ifBLIN_QW0("mife_aopen stdout");
            ex = EX_IOERR;
            goto out;
    }   }
    if  (0 > (ex = csv2pg_do(cf))) ifBLIN_QW0("csv2pg_do");
    if  (!ex && (cf->flags & CVS2PG_WHENY)) {
        fprintf( stderr, "%s\t%d\tfields\n%s\t%d\ttuples\n"
               , cf->inname, cf->fieldnom
               , cf->inname, cf->tuplecount
               );
    }
out:
    exit(ex);
#   undef blin_internal_flags
}
